@@ -1,4 +1,4 @@
-# Pyganim (pyganim.py, ver 0.9.0)
+# Pyganim
 # A sprite animation module for Pygame.
 #
 # By Al Sweigart al@inventwithpython.com
@@ -14,13 +14,14 @@
 
 
 # TODO: Feature idea: if the same image file is specified, re-use the Surface object. (Make this optional though.)
-# TODO: autoplay member for the ctor
-# TODO: sprite sheet support
-# TODO: automatically load animated gifs and pngs.
+# TODO: Dynamically change durations and number of frames.
 
-__version__ = '0.9.0'
+__version__ = '0.9.2'
+
+import pygame
+import time
+import os
 
-import pygame, time
 
 # setting up constants
 PLAYING = 'playing'
@@ -28,41 +29,108 @@ PAUSED = 'paused'
 STOPPED = 'stopped'
 
 # These values are used in the anchor() method.
-NORTHWEST = 'northwest'
-NORTH = 'north'
-NORTHEAST = 'northeast'
-WEST = 'west'
-CENTER = 'center'
-EAST = 'east'
-SOUTHWEST = 'southwest'
-SOUTH = 'south'
-SOUTHEAST = 'southeast'
+NW = NORTHWEST = 'nw'
+N = NORTH = 'n'
+NE = NORTHEAST = 'ne'
+W = WEST = 'w'
+C = CENTER = 'c'
+E = EAST = 'e'
+SW = SOUTHWEST = 'sw'
+S = SOUTH = 's'
+SE = SOUTHEAST = 'se'
+
+TIME_FUNC = lambda: int(time.time() * 1000)
+
+def getImagesFromSpriteSheet(filename, width=None, height=None, rows=None, cols=None, rects=None):
+    """Loads several sprites from a single image file (a "spritesheet").
+
+    One (and only one) of the following parameters should be specified:
+        * width & height of each sprite (all must be the same size)
+        * number of rows and columns of sprites (all must be the same size)
+        * rects, which is a list of tuples formatted as (pygame.Rect, index) or (left, top, width, height)
+    """
+
+    argsType = '' # there should be exactly 1 set of arguments passed (i.e. don't pass width/height AND rows/cols)
+    if (width is not None or height is not None) and (argsType == ''):
+        argsType = 'width/height'
+        assert width is not None and height is not None, 'Both width and height must be specified'
+        assert type(width) == int and width > 0, 'width arg must be a non-zero positive integer'
+        assert type(height) == int and height > 0, 'height arg must be a non-zero positive integer'
+    if (rows is not None or cols is not None) and (argsType == ''):
+        argsType = 'rows/cols'
+        assert rows is not None and cols is not None, 'Both rows and cols must be specified'
+        assert type(rows) == int and rows > 0, 'rows arg must be a non-zero positive integer'
+        assert type(cols) == int and cols > 0, 'cols arg must be a non-zero positive integer'
+    if (rects is not None) and (argsType == ''):
+        argsType = 'rects'
+        for i, rect in enumerate(rects):
+            assert len(rect) == 4, 'rect at index %s is not a sequence of four ints: (left, top, width, height)' % (i)
+
+            assert (type(rect[0]), type(rect[1]), type(rect[2]), type(rect[3])) == (int, int, int, int), 'rect '
+    if argsType == '':
+        raise ValueError('Only pass one set of args: width & height, rows & cols, *or* rects')
+
+    sheetImage = pygame.image.load(filename)
+
+    if argsType == 'width/height':
+        for y in range(0, sheetImage.get_height(), (sheetImage.get_height() // height)):
+            if y + height > sheetImage.get_height():
+                continue
+            for x in range(0, sheetImage.get_width(), (sheetImage.get_width() // width)):
+                if x + width > sheetImage.get_width():
+                    continue
+
+                rects.append((x, y, width, height))
+
+    if argsType == 'rows/cols':
+        spriteWidth = sheetImage.get_width() // cols
+        spriteHeight = sheetImage.get_height() // rows
+
+        for y in range(0, sheetImage.get_height(), spriteHeight):
+            if y + spriteHeight > sheetImage.get_height():
+                continue
+            for x in range(0, sheetImage.get_width(), spriteWidth):
+                if x + spriteWidth > sheetImage.get_width():
+                    continue
+
+                rects.append((x, y, spriteWidth, spriteHeight))
+
+    # create a list of Surface objects from the sprite sheet
+    returnedSurfaces = []
+    for rect in rects:
+        surf = pygame.Surface((rect[2], rect[3]), 0, sheetImage) # create Surface with width/height in rect
+        surf.blit(sheetImage, (0, 0), rect, pygame.BLEND_RGBA_ADD)
+        returnedSurfaces.append(surf)
+
+    return returnedSurfaces
 
 
-class PygAnimation(object):
+
+
+class PygAnimation():
     def __init__(self, frames, loop=True):
         # Constructor function for the animation object. Starts off in the STOPPED state.
         #
         # @param frames
         #     A list of tuples for each frame of animation, in one of the following format:
-        #       (image_of_frame<pygame.Surface>, duration_in_seconds<int>)
-        #       (filename_of_image<str>, duration_in_seconds<int>)
+        #       (image_of_frame<pygame.Surface>, duration_in_milliseconds<int>)
+        #       (filename_of_image<str>, duration_in_milliseconds<int>)
         #     Note that the images and duration cannot be changed. A new PygAnimation object
         #     will have to be created.
         # @param loop Tells the animation object to keep playing in a loop.
 
         # _images stores the pygame.Surface objects of each frame
         self._images = []
-        # _durations stores the durations (in seconds) of each frame.
-        # e.g. [1, 1, 2.5] means the first and second frames last one second,
+        # _durations stores the durations (in milliseconds) of each frame.
+        # e.g. [1000, 1000, 2500] means the first and second frames last one second,
         # and the third frame lasts for two and half seconds.
         self._durations = []
         # _startTimes shows when each frame begins. len(self._startTimes) will
         # always be one more than len(self._images), because the last number
         # will be when the last frame ends, rather than when it starts.
-        # The values are in seconds.
+        # The values are in milliseconds.
         # So self._startTimes[-1] tells you the length of the entire animation.
-        # e.g. if _durations is [1, 1, 2.5], then _startTimes will be [0, 1, 2, 4.5]
+        # e.g. if _durations is [1000, 1000, 2500], then _startTimes will be [0, 1000, 2000, 4500]
         self._startTimes = None
 
         # if the sprites are transformed, the originals are kept in _images
@@ -74,37 +142,44 @@ class PygAnimation(object):
         self._rate = 1.0 # 2.0 means play the animation twice as fast, 0.5 means twice as slow
         self._visibility = True # If False, then nothing is drawn when the blit() methods are called
 
-        self._playingStartTime = 0 # the time that the play() function was last called.
-        self._pausedStartTime = 0 # the time that the pause() function was last called.
+        self._playingStartTime = 0 # the time that the play() function was last called. Epoch timestamp in milliseconds.
+        self._pausedStartTime = 0 # the time that the pause() function was last called. Epoch timestamp in milliseconds.
+
+        # NOTE: There is no "self._elapsed" attribute. "Elapsed" is always calculated based on the current time and _playingStartTime.
+        if type(frames) == str and frames.lower().endswith('.gif'):
+            # frames is an animated gif filename
+            gifImages = splitGif(frames)
+            # NOTE: I can't understand how PIL and Pygame's tobytes()/tostring()/fromstring() methods work, hence the use of temporary files.
+            # NOTE: Named temp files delete themselves on close on Windows, so the temp file has to be created here.
+            frames = []
+            for im in gifImages:
+                im.save('.temp_pyganim.gif')
+                frames.append((pygame.image.load('.temp_pyganim.gif'), im.info['duration'])) # gif duration is already in milliseconds
+                os.unlink('.temp_pyganim.gif')
 
         if frames != '_copy': # ('_copy' is passed for frames by the getCopies() method)
             self.numFrames = len(frames)
             assert self.numFrames > 0, 'Must contain at least one frame.'
             for i in range(self.numFrames):
                 # load each frame of animation into _images
-                frame = frames[i]
-                assert type(frame) in (list, tuple) and len(frame) == 2, 'Frame %s has incorrect format.' % (i)
+                frame = (frames[i][0], int(frames[i][1]))
+                assert type(frame) in (list, tuple) and len(frame) == 2, 'Frame %s has incorrect format. It should be a tuple of length 2, first item a pygame.Surface or image filename, second item an int/float of duration.' % (i)
                 assert type(frame[0]) in (str, pygame.Surface), 'Frame %s image must be a string filename or a pygame.Surface' % (i)
                 assert frame[1] > 0, 'Frame %s duration must be greater than zero.' % (i)
                 if type(frame[0]) == str:
                     frame = (pygame.image.load(frame[0]), frame[1])
                 self._images.append(frame[0])
                 self._durations.append(frame[1])
-            self._startTimes = self._getStartTimes()
 
-
-    def _getStartTimes(self):
-        # Internal method to get the start times based off of the _durations list.
-        # Don't call this method.
-        startTimes = [0]
-        for i in range(self.numFrames):
-            startTimes.append(startTimes[-1] + self._durations[i])
-        return startTimes
+            # calculate start times of each frame
+            self._startTimes = [0]
+            for i in range(self.numFrames):
+                self._startTimes.append(self._startTimes[-1] + self._durations[i])
 
 
     def reverse(self):
-        # Reverses the order of the animations.
-        self.elapsed = self._startTimes[-1] - self.elapsed
+        # Reverses the order of the frames.
+        self.elapsed = (self._durations[-1] + self._startTimes[-1]) - self.elapsed
         self._images.reverse()
         self._transformedImages.reverse()
         self._durations.reverse()
@@ -139,7 +214,7 @@ class PygAnimation(object):
         return retval
 
 
-    def blit(self, destSurface, dest):
+    def blit(self, destSurface, dest=(0, 0)):
         # Draws the appropriate frame of the animation to the destination Surface
         # at the specified position.
         #
@@ -244,14 +319,12 @@ class PygAnimation(object):
         return not self.loop and self.elapsed >= self._startTimes[-1]
 
 
-    def play(self, startTime=None):
+    def play(self, startTime=None): # startTime is epoch timestamp in milliseconds. TODO: Change this?
         # Start playing the animation.
 
         # play() is essentially a setter function for self._state
-        # NOTE: Don't adjust the self.state property, only self._state
-
         if startTime is None:
-            startTime = time.time()
+            startTime = TIME_FUNC()
 
         if self._state == PLAYING:
             if self.isFinished():
@@ -264,6 +337,8 @@ class PygAnimation(object):
         elif self._state == PAUSED:
             # if animation was paused, start playing from where it was paused
             self._playingStartTime = startTime - (self._pausedStartTime - self._playingStartTime)
+        else:
+            assert False, '_state attribute contains an invalid value: %s' % (str(self._state)[:40])
         self._state = PLAYING
 
 
@@ -271,19 +346,19 @@ class PygAnimation(object):
         # Stop having the animation progress, and keep it at the current frame.
 
         # pause() is essentially a setter function for self._state
-        # NOTE: Don't adjust the self.state property, only self._state
-
         if startTime is None:
-            startTime = time.time()
+            startTime = TIME_FUNC()
 
         if self._state == PAUSED:
             return # do nothing
         elif self._state == PLAYING:
             self._pausedStartTime = startTime
         elif self._state == STOPPED:
-            rightNow = time.time()
+            rightNow = TIME_FUNC()
             self._playingStartTime = rightNow
             self._pausedStartTime = rightNow
+        else:
+            assert False, '_state attribute contains an invalid value: %s' % (str(self._state)[:40])
         self._state = PAUSED
 
 
@@ -291,7 +366,7 @@ class PygAnimation(object):
         # Reset the animation to the beginning frame, and do not continue playing
 
         # stop() is essentially a setter function for self._state
-        # NOTE: Don't adjust the self.state property, only self._state
+        assert self._state in (PLAYING, PAUSED, STOPPED), '_state attribute contains an invalid value: %s' % (str(self._state)[:40])
         if self._state == STOPPED:
             return # do nothing
         self._state = STOPPED
@@ -301,22 +376,15 @@ class PygAnimation(object):
         # If paused, start playing. If playing, then pause.
 
         # togglePause() is essentially a setter function for self._state
-        # NOTE: Don't adjust the self.state property, only self._state
-
         if self._state == PLAYING:
-            if self.isFinished():
-                # the one exception: if this animation doesn't loop and it
-                # has finished playing, then toggling the pause will cause
-                # the animation to replay from the beginning.
-                #self._playingStartTime = time.time() # effectively the same as calling play()
-                self.play()
-            else:
-                self.pause()
+            self.pause()
         elif self._state in (PAUSED, STOPPED):
             self.play()
+        else:
+            assert False, '_state attribute contains an invalid value: %s' % (str(self._state)[:40])
 
 
-    def areFramesSameSize(self):
+    def framesAreSameSize(self):
         # Returns True if all the Surface objects in this animation object
         # have the same width and height. Otherwise, returns False
         width, height = self.getFrame(0).get_size()
@@ -355,7 +423,7 @@ class PygAnimation(object):
         # specific "anchor point" (one of the NORTH, SOUTH, SOUTHEAST, etc. constants)
         #
         # By default, they are all anchored to the NORTHWEST corner.
-        if self.areFramesSameSize():
+        if self.framesAreSameSize():
             return # nothing needs to be anchored
             # This check also prevents additional calls to anchor() from doing
             # anything, since anchor() sets all the image to the same size.
@@ -418,20 +486,20 @@ class PygAnimation(object):
         self.currentFrameNum -= int(jump)
 
 
-    def rewind(self, seconds=None):
+    def rewind(self, milliseconds=None):
         # Set the elapsed time back relative to the current elapsed time.
-        if seconds is None:
+        if milliseconds is None:
             self.elapsed = 0.0
         else:
-            self.elapsed -= seconds
+            self.elapsed -= milliseconds
 
 
-    def fastForward(self, seconds=None):
+    def fastForward(self, milliseconds):
         # Set the elapsed time forward relative to the current elapsed time.
-        if seconds is None:
-            self.elapsed = self._startTimes[-1] - 0.00002 # done to compensate for rounding errors
+        if milliseconds is None:
+            self.elapsed = self._startTimes[-1]
         else:
-            self.elapsed += seconds
+            self.elapsed += milliseconds
 
     def _makeTransformedSurfacesIfNeeded(self):
         # Internal-method. Creates the Surface objects for the _transformedImages list.
@@ -573,7 +641,7 @@ class PygAnimation(object):
             # we need to modify the _playingStartTime so that the rest of
             # the animation will play, and then stop. (Otherwise, the
             # animation will immediately stop playing if it has already looped.)
-            self._playingStartTime = time.time() - self.elapsed
+            self._playingStartTime = TIME_FUNC() - self.elapsed
         self._loop = bool(loop)
 
     loop = property(_propGetLoop, _propSetLoop)
@@ -608,17 +676,18 @@ class PygAnimation(object):
 
 
     def _propSetElapsed(self, elapsed):
-        # NOTE: Do to floating point rounding errors, this doesn't work precisely.
-        elapsed += 0.00001 # done to compensate for rounding errors
-        # TODO - I really need to find a better way to handle the floating point thing.
-
         # Set the elapsed time to a specific value.
+        # NOTE: elapsed is in milliseconds
+
+        if self.state == STOPPED:
+            self.state = PAUSED
+
         if self._loop:
             elapsed = elapsed % self._startTimes[-1]
         else:
-            elapsed = getInBetweenValue(0, elapsed, self._startTimes[-1])
+            elapsed = getBoundedValue(0, elapsed, self._startTimes[-1])
 
-        rightNow = time.time()
+        rightNow = TIME_FUNC()
         self._playingStartTime = rightNow - (elapsed * self.rate)
 
         if self.state in (PAUSED, STOPPED):
@@ -627,32 +696,34 @@ class PygAnimation(object):
 
 
     def _propGetElapsed(self):
-        # NOTE: Do to floating point rounding errors, this doesn't work precisely.
-
         # To prevent infinite recursion, don't use the self.state property,
         # just read/set self._state directly because the state getter calls
         # this method.
 
+        # NOTE: Elapsed is in milliseconds, not seconds.
+        assert self._state in (PLAYING, PAUSED, STOPPED), '_state attribute contains an invalid value: %s' % (str(self._state)[:40])
+
         # Find out how long ago the play()/pause() functions were called.
         if self._state == STOPPED:
             # if stopped, then just return 0
             return 0
-
-        if self._state == PLAYING:
+        elif self._state == PLAYING:
             # if playing, then draw the current frame (based on when the animation
             # started playing). If not looping and the animation has gone through
             # all the frames already, then draw the last frame.
-            elapsed = (time.time() - self._playingStartTime) * self.rate
+            elapsed = (TIME_FUNC() - self._playingStartTime) * self.rate
         elif self._state == PAUSED:
             # if paused, then draw the frame that was playing at the time the
             # PygAnimation object was paused
             elapsed = (self._pausedStartTime - self._playingStartTime) * self.rate
+
+        elapsed = int(elapsed)
+
         if self._loop:
             elapsed = elapsed % self._startTimes[-1]
         else:
-            elapsed = getInBetweenValue(0, elapsed, self._startTimes[-1])
-        elapsed += 0.00001 # done to compensate for rounding errors
-        return elapsed
+            elapsed = getBoundedValue(0, elapsed, self._startTimes[-1])
+        return int(elapsed)
 
     elapsed = property(_propGetElapsed, _propSetElapsed)
 
@@ -665,10 +736,13 @@ class PygAnimation(object):
 
     def _propSetCurrentFrameNum(self, frameNum):
         # Change the elapsed time to the beginning of a specific frame.
+        if self._state == STOPPED:
+            self._state = PAUSED # setting the frame num automatically puts it as paused.
+
         if self.loop:
             frameNum = frameNum % len(self._images)
         else:
-            frameNum = getInBetweenValue(0, frameNum, len(self._images)-1)
+            frameNum = getBoundedValue(0, frameNum, len(self._images)-1)
         self.elapsed = self._startTimes[frameNum]
 
     currentFrameNum = property(_propGetCurrentFrameNum, _propSetCurrentFrameNum)
@@ -704,14 +778,14 @@ class PygConductor(object):
 
     def play(self, startTime=None):
         if startTime is None:
-            startTime = time.time()
+            startTime = TIME_FUNC()
 
         for animObj in self._animations:
             animObj.play(startTime)
 
     def pause(self, startTime=None):
         if startTime is None:
-            startTime = time.time()
+            startTime = TIME_FUNC()
 
         for animObj in self._animations:
             animObj.pause(startTime)
@@ -748,7 +822,7 @@ class PygConductor(object):
         for animObj in self._animations:
             animObj.rewind(seconds)
 
-    def fastForward(self, seconds=None):
+    def fastForward(self, seconds):
         for animObj in self._animations:
             animObj.fastForward(seconds)
 
@@ -809,11 +883,14 @@ class PygConductor(object):
             animObj.unlock()
 
 
-def getInBetweenValue(lowerBound, value, upperBound):
+def getBoundedValue(lowerBound, value, upperBound):
     # Returns the value within the bounds of the lower and upper bound parameters.
     # If value is less than lowerBound, then return lowerBound.
     # If value is greater than upperBound, then return upperBound.
     # Otherwise, just return value as it is.
+    if upperBound < lowerBound:
+        upperBound, lowerBound = lowerBound, upperBound
+
     if value < lowerBound:
         return lowerBound
     elif value > upperBound:
@@ -825,9 +902,9 @@ def findStartTime(startTimes, target):
     # With startTimes as a list of sequential numbers and target as a number,
     # returns the index of the number in startTimes that preceeds target.
     #
-    # For example, if startTimes was [0, 2, 4.5, 7.3, 10] and target was 6,
-    # then findStartTime() would return 2. If target was 12, returns 4.
-    assert startTimes[0] == 0
+    # For example, if startTimes was [0, 2000, 4500, 7300, 10000] and target was 6000,
+    # then findStartTime() would return 2. If target was 12000, returns 4.
+    assert startTimes[0] == 0, 'The first value in the start times list should always be 0.'
     lb = 0 # "lb" is lower bound
     ub = len(startTimes) - 1 # "ub" is upper bound
 
@@ -851,3 +928,32 @@ def findStartTime(startTimes, target):
             lb = i
         elif startTimes[i] > target:
             ub = i
+
+
+def splitGif(filename):
+    # Takes a filename of an animated GIF and returns a list of Image objects of each frame.
+    # Requires PIL or Pillow to be installed
+    from PIL import Image
+
+    def iterAnimatedGifFrames(im):
+        # Iterator for frames in an animated GIF.
+        try:
+            i= 0
+            while 1:
+                im.seek(i)
+                imframe = im.copy()
+                if i == 0:
+                    palette = imframe.getpalette()
+                else:
+                    imframe.putpalette(palette)
+                yield imframe
+                i += 1
+        except EOFError:
+            pass
+
+    gifFile = open(str(filename), 'rb')
+    im = Image.open(gifFile) # passing a file-like object so that we can close it and the unit tests won't complain.
+    retVal = list(iterAnimatedGifFrames(im))
+    gifFile.close()
+    return retVal
+
